import dearpygui.dearpygui as dpg
from pathlib import Path
from io import BytesIO
from PIL import Image
from . import image, more
import numpy as np
import zipfile
import json
import copy

#+----------------------------------------------------------------+
#|                           THEME DATA                           |
#+----------------------------------------------------------------+

#=========== ThemeData ===========
#Almacena los datos necesarios para la ejecucion de los themes
class ThemeData():
    def __init__(self):
        self._themes = []
        self._bind_theme = []

    #============== THEMES ==============
    def add_theme(self, theme: dict):
        self._themes.append(theme)

    def get_theme(self):
        return self._themes

    #============== BIND THEME ==============
    def add_bind_theme(self, bind_theme: dict):
        self._bind_theme.append(bind_theme)

    def get_bind_theme(self):
        return self._bind_theme

theme_data = ThemeData()

#+----------------------------------------------------------------+
#|                         EXECUTE THEME                          |
#+----------------------------------------------------------------+

#=========== ExecuteThemes ===========
#Ejecuta las ordenes del theme
class ExecuteThemes():
    def __init__(self):
        self._theme = theme_data.get_theme()
        self._bind_theme = theme_data.get_bind_theme()
        self._theme_manager()

    #==================== _theme_manager ====================
    #Inicia y reparte el proceso de los themes
    def _theme_manager(self):
        #Lista de llaves 
        used_keys = []

        #---------- Bucle que recorre la orden para aplicar los temas ----------
        for element_bind in self._bind_theme:
            #Toma el tag de la orden
            tag = element_bind["tag"]
            #Toma el theme dependiendo del tag_theme asignado en el bind_theme
            themes = copy.deepcopy(next((thems for thems in self._theme if thems['id'] == element_bind["tag_theme"]), None))
            register = image.ImageValidation.create_register(tag, themes)

            key_theme = list(register.keys())[0]  # Guarda la key usada
            used_keys.append(key_theme)
            themes = image.ImageValidation.JSON_validation(register, themes)

            if themes["themes"]:
                self._create_image(themes, tag)

        # Luego de todo, elimina las keys no usadas 
        image.ImageValidation.delete_keys(used_keys)

        self._apply_themes()

    #==================== _create_image ====================
    #Da las ordenes para crear las imagenes
    def _create_image(self, themes, tag):
        #---------- Recorre los temas de las imagenes a crear ----------
        for t in themes["themes"]:
            #Para crear la imagen es necesario tomar el tamaño de los widgets
            width, height = dpg.get_item_rect_size(tag)
            #---------- Realiza las imagenes dependiendo de su tipo ----------
            theme_type = t["theme"]
            match theme_type:
                #---------- Crea la imagen de background gradiente ----------
                case "background_gradiant":
                    #Crea la imagen gradiante
                    img = image.ImageStyleQs.CreateImage.background_gradiants(  gradiant_type=t["gradiant_type"], 
                                                                                width=width,
                                                                                height=height,
                                                                                start_color=t["start_color"],
                                                                                end_color=t["end_color"])
                    #Agrega efectos (rounded_corners)
                    img = image.ImageStyleQs.EffectsImage(img).rounded_corners(t["rounded_corners"])\
                        .image_return()
                    #Guarda la imagen en el zip
                    image.ImageStyleQs.save(img, tag, "core_effects", "background_gradiant")

                #---------- Crea la imagen de background color ----------
                case "background_color":
                    #Crea la imagen de color
                    img = image.ImageStyleQs.CreateImage.backgroud_color(width=width,
                                                                        height= height,
                                                                        color=t["color"])
                    #Agrega efectos (rounded_corners)
                    img = image.ImageStyleQs.EffectsImage(img).rounded_corners(t["rounded_corners"])\
                        .image_return()
                    #Guarda la imagen en el zip
                    image.ImageStyleQs.save(img, tag, "core_effects", "background_color")

                #---------- Crea la imagen de sombreado ----------
                case "shadow":
                    widget_type = themes["widget_type"]
                    match widget_type:
                        #---------- Crea la sombra para los childwindows ----------
                        case "mvChildWindow":
                            #Toma la imgen base(background grandiant o color)
                            try:
                                img = image.ImageStyleQs.load_image(tag, "core_effects", "core.png")
                            except:
                                print(f"The widget with the tag {tag} does not have a base to create the shadow, previously add a background color or gradient for example")
                            else:
                                #Agrega efectos (silhouette_fill, outline, blur, opacity)
                                img = image.ImageStyleQs.EffectsImage(img)\
                                    .silhouette_fill(color=t["color"])\
                                    .outline(color=t["color"], thickness=t["thickness"])\
                                    .blur(t["blur_level"])\
                                    .opacity(t["opacity"])\
                                    .image_return()
                                #Guarda la imagen en el zip
                                image.ImageStyleQs.save(img, tag, "image_layer", "shadow")

                        #---------- Crea la sombra para los widgets de imagenes ----------
                        case "mvImage":
                            #Toma el user data de la imagen(debe ser creada por styleqs)
                            user_data = dpg.get_item_user_data(tag)
                            #Toma la ruta de la imagen
                            img = Image.open(user_data["route"])
                            #Agrega efectos (silhouette_fill, outline, blur, opacity)
                            img = image.ImageStyleQs.EffectsImage(img)\
                                    .silhouette_fill(color=t["color"])\
                                    .outline(color=t["color"], thickness=t["thickness"])\
                                    .blur(t["blur_level"])\
                                    .opacity(t["opacity"])\
                                    .image_return()
                            #Guarda la imagen en el zip
                            image.ImageStyleQs.save(img, tag, "image_layer", "shadow")

    #==================== _register_img_zip ====================
    #Registra la imagen que se encuentra en el zip
    def _register_img_zip(self, router):
        #Consigue la carpeta en donde se encuentra el zip
        base_dir = Path(__file__).resolve().parent.parent
        compressed_images = base_dir / "archives" / "compressed_images.zip"

        #Abre el zip en modo lectura
        with zipfile.ZipFile(compressed_images, "r") as zipimg:
            #Ubica la imagen a utilizar
            with zipimg.open(router) as img_file:
                #Abre la imagen con BytesIO
                image_data = BytesIO(img_file.read())

        #Convierte la imagen en formato RGBA de Pillow
        image_dpg = Image.open(image_data).convert("RGBA")
        #Obtiene el tamaño
        width, height = image_dpg.size
        #Convierte con numpy el formato de pillow a el formato de lectura de dearpygui
        image_np = np.array(image_dpg).flatten() / 255.0

        #Registra la imagen
        with dpg.texture_registry():
            img = dpg.add_static_texture(width, height, image_np)
        return img, width, height

    #==================== _apply_themes ====================
    #Aplica los temas a los widgets
    def _apply_themes(self):
        #Consigue la carpeta en donde se encuentra el JSON
        base_dir = Path(__file__).resolve().parent.parent
        image_registration = base_dir / "data" / "image_registration.json"

        #Abre el JSON en modo de lectura
        with open(image_registration, "r") as archive_JSON:
            #Carga el JSON a la variable content
            content = json.load(archive_JSON)

            #---------- Recorre las llaves (tambien carpetas en el zip y tags de los widgets) del JSON ----------
            for c in content:
                #consigue valores de los widgets (tag, tamaño y pos(posicion))
                tag = c
                width_w, height_w = dpg.get_item_rect_size(c)
                pos = dpg.get_item_pos(c)

                #core_effects(imagen principal, normalmente backgrounds)
                core_effects = content[c]["core_effects"]
                #image_layer(capas de imagenes como pueden ser los sombreados)
                image_layer = content[c]["image_layer"]
                #widget_type(agarra el tipo de widget almacenado en el JSON)
                widget_type = content[c]["widget_type"]

                #Si core_effects no esta vacio significa que hay temas que aplicar
                if core_effects:
                    #Recorre los temas de core
                    for ce in core_effects:
                        #ruta de todos los core en el zip
                        router = f"{tag}/core.png"
                        #Toma la imgen registrada
                        img_ce, _, _ = self._register_img_zip(router)

                        #Adquiere el tema
                        theme_ce = ce["theme"]

                        #---------- Si la palabra "background" esta en el tema se aplica su respectiva configuracion ----------
                        if "background" in theme_ce:
                            children = dpg.get_item_children(tag, 1)
                            #Si existe mas de un widget en el contenedor se aplica antes del primer
                            if children and len(children) > 0:
                                first_child = children[0]
                                dpg.add_image(img_ce, width=width_w, height=height_w, pos=(0,0), parent=tag, before=first_child)
                            #Si no, simplemente se agrega como el primero
                            else:
                                dpg.add_image(img_ce, width=width_w, height=height_w, pos=(0,0), parent=tag)

                #Si image_layer no esta vacio significa que hay temas que aplicar
                if image_layer:
                    #Recorre los temas de image layer
                    for il in image_layer:
                        #ruta de todos las imagen layer en el zip
                        router = f"{tag}/image_layer/{il["theme"]}.png"
                        #Toma la imgen registrada y su tamaño
                        img_il, width, height = self._register_img_zip(router)
                        #Adquiere el tema
                        theme_il = il["theme"]

                        #---------- Agrega casos para los tipos de temas ----------
                        match theme_il:
                            #---------- Aplica el tema shadow(sombreado) ----------
                            case "shadow":
                                match widget_type:
                                    #---------- Caso de aplicacion para los widgets image ----------
                                    case "mvImage":
                                        #Toma el tamaño original del img
                                        width_img, height_img = dpg.get_item_user_data(tag)["size_img"]

                                        #Calcula el tamaño del sombreado dependiendo del tamaño original de la imagen, tamaño del widgets, y tamaño de la imagen shadow
                                        width_shadow = (width_w / width_img) * width
                                        height_shadow = (height_w / height_img) * height

                                        #Posicion del sombreado dado por el usuario
                                        pos_shadow = il["pos"]

                                        #Calcula la posicion
                                        x, y = pos
                                        x -= (width_shadow - width_w) / 2
                                        y -= (height_shadow - height_w) / 2
                                        #Agrega el widget con la imagen de sombreado
                                        dpg.add_image(img_il,   width=width_shadow + pos_shadow[0], 
                                                                height=height_shadow + pos_shadow[1], 
                                                                before=tag, pos=(x, y))

                                    #---------- Caso de aplicacion para los widgets childwindow ----------
                                    case "mvChildWindow":
                                        #Posicion del sombreado dado por el usuario
                                        pos_shadow = il["pos"]

                                        #Calcula la posicion
                                        x, y = pos
                                        x -= (width - width_w) / 2
                                        y -= (height - height_w) / 2

                                        #Agrega el widget con la imagen de sombreado
                                        dpg.add_image(img_il,   width=width + pos_shadow[0], 
                                                                height=height + pos_shadow[1], 
                                                                before=tag, pos=(x, y))
                                        
                more.execute_after()


import dearpygui.dearpygui as dpg
from . import theme
import traceback

#+----------------------------------------------------------------+
#|                          LAYOUT DATA                           |
#+----------------------------------------------------------------+

#=========== LayoutData ===========
#Almacena los datos necesarios para la ejecucion del layout
class LayoutData():
    def __init__(self):
        self._structure_layout = {}
        self._custom_layout = []

    #============ STRUCTURE LAYOUT ============
    def add_structure_layout(self, layout_map):
        self._structure_layout = layout_map

    def get_structure_layout(self):
        return self._structure_layout

    #============== CUSTOM LAYOUT ==============
    def add_custom_layout(self, childwindow_layout: list):
        self._custom_layout.append(childwindow_layout)
    
    def get_custom_layout(self):
        return self._custom_layout
    
    def get_next_custom_layout(self):
        if self._custom_layout:
            return self._custom_layout.pop(0)

layout_data = LayoutData()

#+----------------------------------------------------------------+
#|                        EXTERNAL FUCTION                        |
#+----------------------------------------------------------------+

#=========== wait_rendering ===========
"""
Esta funcion espera el renderizado de los widgets para ejecutar las funcion del layout
si no se espera el renderizado es posible que el layout se realice manera inesperada.
"""
def wait_rendering(tags: list, callback):
    #---------- Validacion de datos de los parametros ----------
    if not isinstance(tags, (str, list)):
        raise ValueError("Parameter 'tags' must be either a string or a list.")

    #Convierte un string a una lista
    if isinstance(tags, str):                                  
        tags = [tags]

    previous_sizes = {tag : None for tag in tags}

    #---------- Verifica el estado de los widgets ----------
    def verify():
        try:
            #---------- Verificando la existencia de todos los tags ----------
            for tag in tags:
                #si aun no existe se vuelve a llamar la funcion "verify" un frame despues
                if not dpg.does_item_exist(tag):
                    dpg.set_frame_callback(dpg.get_frame_count() + 1, verify)
                    return

            all_stable = True

            #---------- Despues de verificar la existencia de los tags, se recorren ----------
            for tag in tags:
                #Toma el tamaño del widget
                current_size = dpg.get_item_rect_size(tag)

                #---------- Si el tamaño es (0,0) aun no se considera terminado la renderizacion ----------
                if current_size == (0,0):
                    all_stable = False
                    break

                previous_size = previous_sizes[tag]

                #---------- Si no se consigue el tamaño o si el tamaño no es igual al anterior no se considera renderizado ----------
                if previous_size is None or current_size != previous_size:
                    previous_sizes[tag] = current_size
                    all_stable = False
                    break

            #---------- Si todos los widgets son estables se ejecuta la funcion dada, si no se vuelve a verificar ----------
            if all_stable:
                callback()
            else:
                dpg.set_frame_callback(dpg.get_frame_count() + 1, verify)

        #---------- Si hay un error lo imprime ----------
        except Exception:
            traceback.print_exc()
            dpg.set_frame_callback(dpg.set_frame_callback(dpg.get_frame_count() + 1, verify))

    #Ejecuta la funcion "verify" despues del frame actual
    dpg.set_frame_callback(dpg.get_frame_count() + 1, verify)

#=========== window_layout_reading ===========
"""
Lee la estructura del window de manera recursiva
pasando por los childwindows
"""
def window_layout_reading(tag: str) -> dict:
    #Contiene la estructura del window
    layout_map = {}

    #=========== _recursively_read ===========
    #Funcion que se utiliza para leer de forma recursiva
    def _recursively_read(current_tag):
        #Toma los hijos
        all_slots = dpg.get_item_children(current_tag)
        #Si hay al menos dos tipos de hijos, entonces toma los del índice 1, si no, deja una lista vacía.
        slot_widgets = all_slots[1] if all_slots and len(all_slots) > 1 else []
        #Crea una lista con el tag del contenedor(window o childwindow)
        layout_map[current_tag] = []

        #---------- Recorreo los widgets hijos ----------
        for w in slot_widgets:
            #Verifica la existencia de un widget
            if not dpg.does_item_exist(w):
                continue                                         

            #Agrega los tags asignados por dearpygui
            layout_map[current_tag].append(w)
            #Obtiene la info del widget
            info = dpg.get_item_info(w)

            #Si el hijo también es un mvChildWindow, llamar recursivamente
            if info["type"] == "mvAppItemType::mvChildWindow":
                _recursively_read(w)

    _recursively_read(current_tag= tag)
    return layout_map

#+----------------------------------------------------------------+
#|                         EXECUTE LAYOUT                         |
#+----------------------------------------------------------------+

#=========== ExecuteLayout ===========
#Ejecuta las ordenes del layout
class ExecuteLayout():
    def __init__(self):
        self._structure_layout = layout_data.get_structure_layout()
        self._custom_layout = layout_data.get_custom_layout()
        self._layout_manager()

    #==================== _layout_manager ====================
    #Inicia y reparte el proceso de los layout
    def _layout_manager(self):
        #---------- Verfica si la lista de custom layout esta vacia, si es asi, termina todo ----------
        if not self._custom_layout:
            theme.ExecuteThemes()
            return

        #Toma la primera orden del custom layout
        next_layout = self._custom_layout.pop(0)
        #Reemplaza el tag por el id que asigna dearpygui
        next_layout["tag"] = dpg.get_alias_id(next_layout["tag"])
        #Toma los widgets hijo del contenedor donde se aplicara el layout
        widgets_layout = self._structure_layout.get(next_layout["tag"])
        #Toman los llaves de next_layout
        parent_tag = next_layout["tag"]
        elements_layout = next_layout["elements_layout"]

        #Convierte todos los tags en id de dearpygui
        for layout in elements_layout:
            layout["tag"] = dpg.get_alias_id(layout["tag"])

        #==================== _main ====================
        #Reparte las tareas de los layout
        def _main():
            match next_layout["type"]:
                case "flex_vertical_layout":
                    self._apply_flex_vertical_layout(widgets_layout, next_layout, parent_tag)
                case "flex_horzizontal_layout":
                    self._apply_flex_horizontal_layout(widgets_layout, next_layout, parent_tag)
                case "grid_layout":
                    self._apply_grid_layout(widgets_layout, next_layout, parent_tag)

        #Espera el renderizado de los widgets y ejecuta la funcion "_main"
        wait_rendering(widgets_layout, _main)

    #==================== _calculate_size ====================
    #Calcula el ancho y alto de los contenedores(childWindow, Window)
    def _calculate_size(self, type_size, calculate):
        #Tag del contenedor
        parent_tag = calculate["parent_tag"]
        #Esta variable almacena un valor de 0 o 1 es decir agarra en size el 0(width) y 1(height)
        w_h_int = calculate["w_h_int"]

        #---------- Deja el valor actual del contenedor ----------
        if type_size == "default":
            size = dpg.get_item_rect_size(parent_tag)
            return size[w_h_int]

        #---------- Calcula el tamaño dependiendo de los widgets (existen 2 tipos) ----------
        elif type_size == "auto":
            match calculate["calculate_type"]:
                #---------- Toma el tamaño recorriendo los widgets y agregando el gap ----------
                case "cycle_through_widgets":
                    parameters = calculate["calculation_type_parameters"]

                    widgets_layout = parameters["widgets_layout"]
                    padding = parameters["padding"]
                    gap = parameters["gap"]

                    w_or_h = padding
                    for widget in widgets_layout:
                        size_widget = dpg.get_item_rect_size(widget)
                        width_or_height = size_widget[w_h_int]
                        w_or_h += width_or_height + gap
                    return w_or_h

                #---------- Toma el tamaño dependiendo del widget con mas tamaño ----------
                case "get_maximum":
                    parameters = calculate["calculation_type_parameters"]

                    widgets_layout = parameters["widgets_layout"]
                    padding = parameters["padding"]

                    w_or_h = 0
                    for widget in widgets_layout:
                        size_widget = dpg.get_item_rect_size(widget)
                        w_or_h = max(w_or_h, size_widget[w_h_int])
                    return w_or_h + (padding * 2)

    #==================== _pos_absolute ====================
    #Agrega la posicion absoluta a los widgets, tiene diversas opciones por defecto
    def _pos_absolute(self, position_absolute, parent_width, parent_height, width, height):
        match position_absolute:
            case "top":
                return ((parent_width / 2) - (width / 2) , 0)
            case "bottom":
                return ((parent_width / 2) - (width / 2) , parent_height - height)
            case "left":
                return (0 , (parent_height / 2) - (height / 2))
            case "right":
                return (parent_width - width , (parent_height / 2) - (height / 2))
            case "center":
                return ((parent_width / 2) - (width / 2) , (parent_height / 2) - (height / 2))
            case "top_left":
                return (0, 0)
            case "top_right":
                return (parent_width - width, 0)
            case "bottom_left":
                return (0, parent_height - height)
            case "bottom_right":
                return (parent_width - width, parent_height - height)
            case (int(x), int(y)):
                return (x, y)

    #==================== _pos_relative ====================
    #<<<<<<<<<<<<<<<<<<< AUN NO AGREGADO >>>>>>>>>>>>>>>>>>>
    def _pos_relative(self):
        print("The _pos_relative function is not yet available.")

    #==================== _apply_flex_vertical_layout ====================
    #realiza el proceso de aplicacion del flex layout de manera vertical
    def _apply_flex_vertical_layout(self, widgets_layout, layout_data, parent_tag):
        #Toma los elementos (widgets) que tiene layout personalizado
        elements_layout = layout_data["elements_layout"]
        #Toma los parametros del contenedor(widget padre o childwindow, window)
        layout_parent = layout_data["layout"]

        #Variables con los elementos de layout_parent
        width_layout = layout_parent["width"]
        height_layout = layout_parent["height"]
        gap = layout_parent["gap"]
        padding = layout_parent["padding"]
        alling_widgets = layout_parent["alling_widgets"]
        no_scrollbar = layout_parent["no_scrollbar"]

        #---------- Consigue el ancho del padre ----------
        if isinstance(width_layout, int):
            #Si es int significa que el usuario agrego un valor
            width_parent = width_layout
        else:
            #Si no lo es, debe ser "auto" o "default" se llama la funcion que lo calcula
            width_parent = self._calculate_size(width_layout, { "parent_tag": parent_tag,
                                                                "w_h_int": 0,
                                                                "calculate_type": "get_maximum",
                                                                "calculation_type_parameters": {
                                                                    "widgets_layout": widgets_layout,
                                                                    "padding": padding
                                                                }})

        #---------- Consigue el alto del padre ----------
        if isinstance(height_layout, int):
            #Si es int significa que el usuario agrego un valor
            height_parent = height_layout
        else:
            #Si no lo es, debe ser "auto" o "default" se llama la funcion que lo calcula
            height_parent = self._calculate_size(height_layout, { "parent_tag": parent_tag,
                                                                "w_h_int": 1,
                                                                "calculate_type": "cycle_through_widgets",
                                                                "calculation_type_parameters": {
                                                                    "widgets_layout": widgets_layout,
                                                                    "padding": padding,
                                                                    "gap": gap
                                                                }})

        #Edita los parametros del padre
        dpg.configure_item( parent_tag,
                            width=width_parent,
                            height=height_parent,
                            no_scrollbar=no_scrollbar,
                            no_scroll_with_mouse=no_scrollbar)

        #Posicion inicial
        y_pos = padding

        #---------- Bucle que recorre todos los widgets y los organiza ----------
        for widget_id in widgets_layout:
            #Tamaño del widget
            width, height = dpg.get_item_rect_size(widget_id)
            #Toma la confirugacion con su tag
            config = next((el for el in elements_layout if el.get("tag") == widget_id), None)
            #posicion inicial del widget
            pos_widget = (0,0)

            #---------- Si configuracion no es None, significa que se agregaron valores modificables ----------
            if config:
                #Varibles que toman el tamaño
                width_widget = config["width"]
                height_widget = config["height"]
                #Toma el tipo de posicionamiento
                position = config["position"]

                #---------- Cambia el ancho del widget ----------
                #Si no es "user" se calcula su valor "percentage" o "fixed", si lo es deja su valor
                if width_widget != "user":
                    size_type = width_widget.size_type
                    #Calcula ancho con el porcentaje 
                    if size_type == "percentage":
                        width = (width_widget.value / 100) * width_parent - (padding * 2)
                        dpg.configure_item(widget_id, width=width)

                    #Ingresa el valor fixed agregado
                    elif size_type == "fixed":
                        width = width_widget.pixels
                        dpg.configure_item(widget_id, width=width)
                    width, height = dpg.get_item_rect_size(widget_id)

                #---------- Cambia el alto del widget ----------
                #Si no es "user" se calcula su valor "percentage" o "fixed", si lo es deja su valor
                if height_widget != "user":
                    size_type = height_widget.size_type

                    #<<<<<<<<<<<<<<<<<<< AUN NO AGREGADO >>>>>>>>>>>>>>>>>>>
                    if size_type == "percentage":
                            print("The application of the percentage with the height in a vertical flex layout is not yet available.")

                    #Ingresa el valor fixed agregado
                    elif size_type == "fixed":
                        height = height_widget.pixels
                        dpg.configure_item(widget_id, height=height)
                    width, height = dpg.get_item_rect_size(widget_id)

                #---------- Aplica el tipo de posicionamiento ----------
                if position.position_type == "static":
                    type_static = position.position_static
                    match type_static:
                        case "start":
                            pos_widget = (padding, y_pos)
                        case "center":
                            pos_widget = ((width_parent / 2) - (width / 2), y_pos)
                        case "end":
                            pos_widget = ((width_parent - width) - padding, y_pos)

            #---------- Si no agrego configuracion del widget, se le agrega el posicionamiento de alling ----------
            else:
                match alling_widgets:
                    case "alling_start":
                        pos_widget = (padding, y_pos)
                    case "alling_center":
                        pos_widget = ((width_parent / 2) - (width / 2), y_pos)
                    case "alling_end":
                        pos_widget = ((width_parent - width) - padding, y_pos)

            #---------- Verifica si se agrego un tipo de posicion en especifico (se agrego aqui por el tipo de posicion relativa) ----------
            if config:
                position = config["position"]
                if position.position_type == "relative":
                    pos_widget = self._pos_relative()
                elif position.position_type == "absolute":
                    pos_widget = self._pos_absolute(position.position_absolute, 
                                                    width_parent, 
                                                    height_parent ,
                                                    width, 
                                                    height)

            #aplica el posicionamiento
            dpg.set_item_pos(widget_id, pos_widget)

            #---------- Identifica si es el ultimo widget para asi no agregarle gap ----------
            if widget_id == widgets_layout[-1]:
                y_pos += height
            else:
                y_pos += height + gap

        #Ejecuta la siguiente orden
        dpg.set_frame_callback(dpg.get_frame_count() + 1, self._layout_manager)

    #==================== _apply_flex_horizontal_layout ====================
    #realiza el proceso de aplicacion del flex layout de manera horizontal
    def _apply_flex_horizontal_layout(self, widgets_layout, layout_data, parent_tag):
        #Toma los elementos (widgets) que tiene layout personalizado
        elements_layout = layout_data["elements_layout"]
        #Toma los parametros del contenedor(widget padre o childwindow, window)
        layout_parent = layout_data["layout"]

        #Variables con los elementos de layout_parent
        width_layout = layout_parent["width"]
        height_layout = layout_parent["height"]
        gap = layout_parent["gap"]
        padding = layout_parent["padding"]
        alling_widgets = layout_parent["alling_widgets"]
        no_scrollbar = layout_parent["no_scrollbar"]

        #---------- Consigue el ancho del padre ----------
        if isinstance(width_layout, int):
            #Si es int significa que el usuario agrego un valor
            width_parent = width_layout
        else:
            #Si no lo es, debe ser "auto" o "default" se llama la funcion que lo calcula
            width_parent = self._calculate_size(width_layout, { "parent_tag": parent_tag,
                                                                "w_h_int": 0,
                                                                "calculate_type": "cycle_through_widgets",
                                                                "calculation_type_parameters": {
                                                                    "widgets_layout": widgets_layout,
                                                                    "padding": padding,
                                                                    "gap": gap
                                                                }})

        #---------- Consigue el alto del padre ----------
        if isinstance(height_layout, int):
            #Si es int significa que el usuario agrego un valor
            height_parent = height_layout
        else:
            #Si no lo es, debe ser "auto" o "default" se llama la funcion que lo calcula
            height_parent = self._calculate_size(height_layout, { "parent_tag": parent_tag,
                                                                "w_h_int": 1,
                                                                "calculate_type": "get_maximum",
                                                                "calculation_type_parameters": {
                                                                    "widgets_layout": widgets_layout,
                                                                    "padding": padding,
                                                                }})

        #Posicion inicial
        x_pos = padding

        #Edita los parametros del padre
        dpg.configure_item( parent_tag,
                            width=width_parent,
                            height=height_parent,
                            no_scrollbar=no_scrollbar,
                            no_scroll_with_mouse=no_scrollbar)

        #---------- Bucle que recorre todos los widgets y los organiza ----------
        for widget_id in widgets_layout:
            #Tamaño del widget
            width, height = dpg.get_item_rect_size(widget_id)
            #Toma la confirugacion con su tag
            config = next ((cfg for cfg in elements_layout if cfg.get("tag") == widget_id), None)
            #Posicion inicial del widget
            pos_widget = (0,0)

            #---------- Si configuracion no es None, significa que se agregaron valores modificables ----------
            if config:
                #Varibles que toman el tamaño
                width_widget = config["width"]
                height_widget = config["height"]
                #Toma el tipo de posicionamiento
                position = config["position"]

                #---------- Cambia el alto del widget ----------
                #Si no es "user" se calcula su valor "percentage" o "fixed", si lo es deja su valor
                if height_widget != "user":
                    size_type = height_widget.size_type

                    #Calcula el alto con el percentage
                    if size_type == "percentage":
                        height = int((height_widget.value / 100) * height_parent - (padding * 2))
                        dpg.configure_item(widget_id, height=height)
                    
                    #Ingresa el valor fixed agregado
                    elif size_type == "fixed":
                        height = height_widget.pixels
                        dpg.configure_item(widget_id, height=height)
                    width, height = dpg.get_item_rect_size(widget_id)

                #---------- Cambia el ancho del widget ----------
                #Si no es "user" se calcula su valor "percentage" o "fixed", si lo es deja su valor
                if width_widget != "user":
                    size_type = width_widget.size_type

                    #<<<<<<<<<<<<<<<<<<< AUN NO AGREGADO >>>>>>>>>>>>>>>>>>>
                    if size_type == "percentage":
                        print("The application of the percentage with the width in a horizontal flex layout is not yet available.")

                    #Ingresa el valor fixed agregado
                    elif size_type == "fixed":
                        width = width_widget.pixels
                        dpg.configure_item(widget_id, width=width)
                    width, height = dpg.get_item_rect_size(widget_id)

                #---------- Aplica el tipo de posicionamiento ----------
                if position.position_type == "static":
                    type_static = position.position_static
                    match type_static:
                        case "start":
                            pos_widget = (x_pos, padding)
                        case "center":
                            pos_widget = (x_pos, (height_parent / 2) - (height / 2))
                        case "end":
                            pos_widget = (x_pos, (height_parent - height) - padding)

            #---------- Si no agrego configuracion del widget, se le agrega el posicionamiento de alling ----------
            else:
                match alling_widgets:
                    case "alling_start":
                        pos_widget = (x_pos, padding)
                    case "alling_center":
                        pos_widget = (x_pos, (height_parent / 2) - (height / 2))
                    case "alling_end":
                        pos_widget = (x_pos, (height_parent - height) - padding)

            #---------- Verifica si se agrego un tipo de posicion en especifico (se agrego aqui por el tipo de posicion relativa) ----------
            if config:
                position = config["position"]
                if position.position_type == "relative":
                    pos_widget = self._pos_relative()
                elif position.position_type == "absolute":
                    pos_widget = self._pos_absolute(position.position_absolute, 
                                                    width_parent, 
                                                    height_parent,
                                                    width, 
                                                    height)

            #aplica el posicionamiento
            dpg.set_item_pos(widget_id, pos_widget)

            #---------- Identifica si es el ultimo widget para asi no agregarle gap ----------
            if widget_id == widgets_layout[-1]:
                x_pos += width
            else:
                x_pos += width + gap

        #Ejecuta la siguiente orden
        dpg.set_frame_callback(dpg.get_frame_count() + 1, self._layout_manager)

    #==================== _apply_grid_layout ====================
    #realiza el proceso de aplicacion del grid layout
    def _apply_grid_layout(self, widgets_layout, layout_data, parent_tag):
        #Tipos de widgets que no permiten modificar el alto
        widgets_without_height = {  "mvAppItemType::mvText", "mvAppItemType::mvCheckbox",
                                    "mvAppItemType::mvRadioButton", "mvAppItemType::mvSelectable",
                                    "mvAppItemType::mvMenuItem", "mvAppItemType::mvListbox",
                                    "mvAppItemType::mvCombo", "mvAppItemType::mvTabButton",
                                    "mvAppItemType::mvTab", "mvAppItemType::mvCollapsingHeader",
                                    "mvAppItemType::mvTreeNode"}
        #Toma los elementos (widgets) que tiene layout personalizado
        elements_layout = layout_data["elements_layout"]
        #Toma los parametros del contenedor(widget padre o childwindow, window)
        layout_parent = layout_data["layout"]
        #Toma los parametros que definen el ancho y alto de las filas o columnas
        define_rows_columns = layout_data["define_rows_columns"]

        #Variables con los elementos de layout_parent
        rows = layout_parent["rows"]
        columns = layout_parent["columns"]
        gap = layout_parent["gap"]
        padding = layout_parent["padding"]
        no_scrollbar = layout_parent["no_scrollbar"]

        #consigue el total de espacios en el grid
        total_space_grid = (rows * columns) + 1

        #Variables que importantes
        #Almacena la ubicacion de los widgets
        place_widgets = []
        #Almacena el ancho y alto de las columnas y filas
        defines = []

        #---------- almacena en un list las columnas y filas definidas ----------
        for define in define_rows_columns:
            define_type = define.define
            match define_type:
                case "row":
                    defines.append({"height": define.height.pixels,
                                    "row": define.row})
                case "column":
                    defines.append({"width": define.width.pixels,
                                    "column": define.column})

        #Variable de columnas y filas que no estan disponibles
        undefined_columns = columns
        undefined_rows = rows

        #---------- Toma el ancho y alto disponible, quitando el padding y gap ----------
        width_space, height_space = dpg.get_item_rect_size(parent_tag)
        available_width_space = (width_space - (padding * 2)) - ((gap - 1) * columns)
        available_height_space = (height_space - (padding * 2)) - ((gap - 1) * rows)

        #---------- Recta el ancho que ya fue definido anteriormente ----------
        unavailable_width_space = next((item for item in defines if item.get("column")), None)
        if unavailable_width_space:
            available_width_space -= unavailable_width_space["width"]
            undefined_columns -= 1

        #---------- Recta el alto que ya fue definido anteriormente ----------
        unavailable_height_space = next((item for item in defines if item.get("row")), None)
        if unavailable_height_space:
            available_height_space -= unavailable_height_space["height"]
            undefined_rows -= 1

        #---------- divide el espacio disponible para las columnas y filas que no fueron definidas ----------
        available_width_space /= undefined_columns
        available_height_space /= undefined_rows

        #Almacena el espacio y las ubicaciones en el eje XY
        space_column = 1
        space_row = 1
        x_pos = padding
        y_pos = padding

        #---------- Bucle que asigna los valores en place_widgets ----------
        for space_grid in range(total_space_grid):
            if space_grid == 0:
                pass
            else:
                #---------- maneja el ancho de los widgets ----------
                width = available_width_space
                current_column_width = next((item for item in defines if item.get("column") == space_column), None)
                if current_column_width:
                    width = current_column_width["width"]

                #---------- maneja el bajado de espacio de widgets ----------
                if (space_grid - 1) % rows == 0 and space_grid != 1:
                    space_column = 1
                    space_row += 1
                    #agrega la posicion de manera vertical
                    y_pos += height + gap
                    #se reinicia el x_pos(posicionamiento de manera vertical) al bajar
                    x_pos = padding

                #---------- maneja el ancho alto de los widgets ----------
                height = available_height_space
                current_row_height = next((item for item in defines if item.get("row") == space_row), None)
                if current_row_height:
                    if current_row_height["row"] == space_row:
                        height = current_row_height["height"]

                #Crea la estructura de los datos
                structure = {space_grid:{"pos": (x_pos, y_pos),
                                        "width": width,
                                        "height": height}}

                #Agrega la posicion de manera horizontal
                x_pos += width + gap
                #Agrega space column
                space_column += 1
                #Agrega la estructura a place_widgets
                place_widgets.append(structure)

        #---------- Agrega los elementos que ya habian definido su posicion ----------
        for element in elements_layout:
            element_place_widget = {}
            for place_w in place_widgets:
                if element["pos"] in place_w:
                    element_x = place_widgets.pop(element["pos"] - 1)
                    element_place_widget = element_x.get(element["pos"])
            #---------- Aplica el tamaño dependiendo del widget ----------
            widget_type = dpg.get_item_type(widget)
            #agrega el elemento con su pos y tamaño
            if widget_type in widgets_without_height:
                dpg.configure_item( element["tag"], 
                                    pos=element_place_widget["pos"], 
                                    width=element_place_widget["width"])
            else:
                dpg.configure_item( element["tag"],
                                    pos=element_place_widget["pos"],
                                    width=element_place_widget["width"],
                                    height=element_place_widget["height"])

            widgets_layout.remove(element["tag"])

        #---------- Agrega los elementos que no han sido definidos ----------
        for widget, place in zip(widgets_layout, place_widgets):
            _, element_place_widget = next(iter(place.items()))

            #---------- Aplica el tamaño dependiendo del widget ----------
            widget_type = dpg.get_item_type(widget)
            #Hay widgets como el texto que no se le puede agregar un alto y su tamaño depende de la fuente
            if widget_type in widgets_without_height:
                dpg.configure_item( widget, 
                                pos=element_place_widget["pos"], 
                                width=element_place_widget["width"])
            else:
                dpg.configure_item( widget, 
                                    pos=element_place_widget["pos"], 
                                    width=element_place_widget["width"], 
                                    height=element_place_widget["height"])

        #Ejecuta la siguiente orden
        dpg.set_frame_callback(dpg.get_frame_count() + 1, self._layout_manager)


from typing import Literal, Union, List, Tuple
import dearpygui.dearpygui as dpg
import traceback

#+----------------------------------------------------------------+
#|                           MORE DATA                            |
#+----------------------------------------------------------------+

#=========== MoreData ===========
#Almacena los datos necesarios para la ejecucion del layout
class MoreData():
    def __init__(self):
        self._execute_after = []

    #============ EXECUTE AFTER ============
    def add_execute_after(self, callback):
        self._execute_after.append(callback)

    def get_execute_after(self):
        return self._execute_after

more_data = MoreData()

#+----------------------------------------------------------------+
#|                             MORE                               |
#+----------------------------------------------------------------+

def execute_after():
    callbacks = more_data.get_execute_after()
    for call in callbacks:
        call()


from typing import Literal, Union, List, Tuple
from .internal import layout, theme, more, add
from dataclasses import dataclass
import dearpygui.dearpygui as dpg
from pathlib import Path
import traceback

#+----------------------------------------------------------------+
#|                           DATA CLASS                           |
#+----------------------------------------------------------------+

#============ POSITION ============
@dataclass
class StaticPosition:
    position_static: Literal["start", "center", "end"] = "start"
    position_type = "static"

    def __post_init__(self):
        if self.position_static not in {"start", "center", "end"}:
            raise ValueError("Parameter 'position_static' must be one of: 'start', 'center', or 'end'.")

@dataclass
class RelativePosition:
    position_relative: Tuple[int, int] = (0,0)
    position_type = "relative"

    def __post_init__(self):
        if not isinstance(self.position_relative, tuple) or len(self.position_relative) != 2:
            raise ValueError("Parameter 'position_relative' must be a tuple of exactly 2 integers (x, y).")
        if not all(isinstance(val, int) for val in self.position_relative):
            raise ValueError("Parameter 'position_relative' must be a tuple of exactly 2 integers (x, y).")

@dataclass
class AbsolutePosition:
    position_absolute:  Union[Tuple[int, int], 
                        Literal["top", "bottom", "left", "right", "center",
                                "top_left", "top_right", "bottom_left", "bottom_right"]] = (0,0)
    position_type = "absolute"

    def __post_init__(self):
        if isinstance(self.position_absolute, tuple):
            if len(self.position_absolute) != 2 or not all(isinstance(val, int) for val in self.position_absolute):
                raise ValueError("Parameter 'position_absolute' must be a tuple of exactly 2 integers (x, y).")
        elif isinstance(self.position_absolute, str):
            if self.position_absolute not in {  "top", "bottom", "left", "right", "center",
                                                "top_left", "top_right", "bottom_left", "bottom_right"}:
                raise ValueError(
                    "Parameter 'position_absolute' must be one of: "
                    "'top', 'bottom', 'left', 'right', 'center', "
                    "'top_left', 'top_right', 'bottom_left', or 'bottom_right'."
                )
        else:
            raise ValueError("Parameter 'position_absolute' must be either a tuple of 2 integers or a valid position string.")

#============== SIZE ==============
@dataclass
class FixedSize:
    pixels: int
    size_type: Literal["fixed"] = "fixed"

    def __post_init__(self):
        if not (1 <= self.pixels):
            raise ValueError("Parameter 'pixels' must be an integer greater than or equal to 1.")

@dataclass
class PercentageSize:
    value: int
    size_type: Literal["percentage"] = "percentage"

    def __post_init__(self):
        if not (1 <= self.value <= 100):
            raise ValueError("Parameter 'value' must be an integer between 1 and 100 (inclusive).")

@dataclass
class GridSize:
    space: int
    size_type: Literal["grid"] = "grid"

    def __post_init__(self):
        if not (1 <= self.space):
            raise ValueError("Parameter 'space' must be an integer greater than or equal to 1.")

#=========== GRID TYPES ===========
@dataclass
class DefineRow:
    row: int
    height: FixedSize
    define: str = "row"

    def __post_init__(self):
        if self.row < 1:
            raise ValueError("Parameter 'row' must be an integer greater than or equal to 1.")
        if not isinstance(self.height, FixedSize):
            raise ValueError("Parameter 'height' must be an instance of FixedSize.")

@dataclass
class DefineColumn:
    column: int
    width: FixedSize
    define: str = "column"

    def __post_init__(self):
        if self.column < 1:
            raise ValueError("Parameter 'column' must be an integer greater than or equal to 1.")
        if not isinstance(self.width, FixedSize):
            raise ValueError("Parameter 'width' must be an instance of FixedSize.")

#+----------------------------------------------------------------+
#|                             LAYOUT                             |
#+----------------------------------------------------------------+

#=========== WindowLayout ===========
"""
Clase principal para el proceso de ejecucion de los layout
Esta clase da la orden de ejecucion de los layout

IMPORTANTE: 
CASO 1: las clases como FlexLayoutBuilder y GridLayoutBuilder
deben de ser ejecutados dentro del contexto de esta clase, si no lo hacen las ordenes
seran ignoradas.

CASO 2: Esta clase inicia el proceso de ejecucion de todo el funcionamiento del modulo
es decir que si no se utiliza, las ordenes del ThemeWidgets y algunas cosas del funcionamiento
de add widget no podrian funcionar, por que? se tomo esta decision por que los modulos como
theme se basan en agregar otro widget y como este modulo utiliza la posicion absoluta para el orden
de los widgets, si no se utiliza, los widgets agregados por las otras funciones o clases no se agregarian como
deberian, por lo cual no tiene sentido ejecutarlos si no se utiliza el layout.
"""
class WindowLayout():
    def __init__(self, tag: Union[str, int]):
        #Validacion
        if not isinstance(tag, (str, int)):
            raise TypeError("Parameter 'tag' must be a string or an integer.")

        self._tag = dpg.get_alias_id(tag)

    def __enter__(self):
        #Agrega la estructura de todos los widgets
        layout.layout_data.add_structure_layout(layout.window_layout_reading(self._tag))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        #Ejecuta la funcion que agrega el nuevo layout
        else : layout.ExecuteLayout()


#=========== FlexLayoutBuilder ===========
"""
Da la orden del flex layout, ordena los widgets de manera horizontal y vertical
tambien maneja los tamaños de los widgets y de su contenedor
"""
class FlexLayoutBuilder():
    def __init__(self, *,
                tag: Union[str, int],
                height: Union[Literal["auto", "default"], int] = "auto",
                width: Union[Literal["auto", "default"], int] = "auto",
                horizontal: bool = False,
                gap: int = 0,
                padding: Union[int, Tuple[int, int], Tuple[int, int, int, int]] = 0,
                alling_widgets: Literal["alling_start", "alling_center", "alling_end"] = "alling_start",
                no_scrollbar: bool = True
                ) -> None:

        #---------- Validacion de datos de los parametros ----------
        if not isinstance(tag, (str, int)):
            raise TypeError("Parameter 'tag' must be a string or an integer.")

        if isinstance(height, str):
            if height not in {"auto", "default"}:
                raise ValueError("Parameter 'height' must be either 'auto', 'default', or a positive integer.")
        elif isinstance(height, int):
            if height <= 0:
                raise ValueError("Parameter 'height' must be a positive integer.")
        else:
            raise ValueError("Parameter 'height' must be either 'auto', 'default', or a positive integer.")
        
        if isinstance(width, str):
            if width not in {"auto", "default"}:
                raise ValueError("Parameter 'width' must be either 'auto', 'default', or a positive integer.")
        elif isinstance(width, int):
            if width <= 0:
                raise ValueError("Parameter 'width' must be a positive integer.")
        else:
            raise ValueError("Parameter 'width' must be either 'auto', 'default', or a positive integer.")

        if not isinstance(horizontal, bool):
            raise ValueError("Parameter 'horizontal' must be a boolean value (True or False).")

        if not isinstance(gap, int) or gap < 0:
            raise ValueError("Parameter 'gap' must be a non-negative integer.")

        if not  (isinstance(padding, int) or 
                (isinstance(padding, tuple) and len(padding) in {2, 4} and all(isinstance(x, int) for x in padding))):
            raise ValueError("Parameter 'padding' must be either an integer, or a tuple of 2 or 4 integers.")

        if alling_widgets not in {"alling_start", "alling_center", "alling_end"}:
            raise ValueError("Parameter 'alling_widgets' must be one of: 'alling_start', 'alling_center', or 'alling_end'.")

        if not isinstance(no_scrollbar, bool):
            raise ValueError("Parameter 'no_scrollbar' must be a boolean value (True or False).")

        #Indica el tipo de layout
        type_layout = "flex_vertical_layout"
        if horizontal:
            type_layout = "flex_horzizontal_layout"

        #Estructura de flex layout
        self._structure_layout = {  "type": type_layout,
                                    "tag": tag,
                                    "layout": { "width": width,
                                                "height": height,
                                                "gap": gap,
                                                "padding": padding,
                                                "alling_widgets": alling_widgets,
                                                "no_scrollbar": no_scrollbar},
                                    "elements_layout": []}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        #Ejecuta la funcion que agrega la nueva instruccion al custom layout
        else : 
            layout.layout_data.add_custom_layout(self._structure_layout)

    #Funcion que agrega diseño unico a widget indicado
    def add_widget( self, *, tag: Union[str, int], 
                    position: Union[StaticPosition, RelativePosition, AbsolutePosition] = StaticPosition(),
                    width: Union[FixedSize, PercentageSize, Literal["user"]] = "user",
                    height: Union[FixedSize, PercentageSize, Literal["user"]] = "user"):

        #---------- Validacion de datos de los parametros ----------
        if not isinstance(tag, (str, int)):
            raise TypeError("Parameter 'tag' must be a string or an integer.")

        if not isinstance(position, (StaticPosition, RelativePosition, AbsolutePosition)):
            raise ValueError("Parameter 'position' must be an instance of StaticPosition, RelativePosition, or AbsolutePosition.")

        if not all(isinstance(val, (FixedSize, PercentageSize, str)) for val in (width, height)):
            raise ValueError("Parameters 'width' and 'height' must each be an instance of FixedSize, PercentageSize, or a string.")

        #Toma los datos en un dict y los envia a "self._structure_layout["elements_layout"]"
        self._structure_layout["elements_layout"].append({  "tag": tag, 
                                                            "position": position, 
                                                            "width": width, "height": height})

#=========== GridLayoutBuilder ===========
"""
Da la orden del grid layout, maneja el orden de los widgets como si fueran una cuadricula
se puede modificar las cuadriculas indicando el tamaño de la fila o columna
los widgets adquieren el tamaño de las cuadriculas que fueron asignadas
"""
class GridLayoutBuilder():
    def __init__(self, *,
                tag: Union[str, int],
                rows: Union[int, None] = None,
                columns: int = 1,
                gap: int = 0,
                padding: Union[int, Tuple[int, int], Tuple[int, int, int, int]] = 0,
                no_scrollbar: bool = True
                ) -> None:

        #---------- Validacion de datos de los parametros ----------
        if not isinstance(tag, (str, int)):
            raise TypeError("Parameter 'tag' must be a string or an integer.")
        
        if rows is not None and (not isinstance(rows, int) or rows < 1):
            raise ValueError("Parameter 'rows' must be an integer greater than or equal to 1, or None.")
        
        if not isinstance(columns, int) or columns < 1:
            raise ValueError("Parameter 'columns' must be an integer greater than or equal to 1.")
        
        if not isinstance(gap, int) or gap < 0:
            raise ValueError("Parameter 'gap' must be a non-negative integer.")
        
        if not  (isinstance(padding, int) or 
                (isinstance(padding, tuple) and len(padding) in {2, 4} and all(isinstance(x, int) for x in padding))):
            raise ValueError("Parameter 'padding' must be either an integer, or a tuple of 2 or 4 integers.")
        
        if not isinstance(no_scrollbar, bool):
            raise ValueError("Parameter 'no_scrollbar' must be a boolean value (True or False).")

        #Estructura de grid layout
        self._structure_layout = {  "type": "grid_layout",
                                    "tag": tag,
                                    "layout": { "rows": rows,
                                                "columns": columns,
                                                "gap": gap,
                                                "padding": padding,
                                                "no_scrollbar": no_scrollbar},
                                    "define_rows_columns": [],
                                    "elements_layout": []}
        
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        #Ejecuta la funcion que agrega la nueva instruccion al custom layout
        else : layout.layout_data.add_custom_layout(self._structure_layout)
    
    #Esta funcion define un tamaño para las columnas y filas
    def define(self, define: Union[DefineColumn, DefineRow]):

        #---------- Validacion de datos de los parametros ----------
        if not isinstance(define, (DefineColumn, DefineRow)):
            raise ValueError("Parameter 'define' must be an instance of either DefineColumn or DefineRow.")

        #Toma los datos de "define" y los envia a "self._structure_layout["define_row_columns"]"
        self._structure_layout["define_rows_columns"].append(define)

    #Funcion que agrega diseño unico a widget indicado
    def place_widget(self, *,
                    tag: Union[str, int],
                    pos: int):

        #---------- Validacion de datos de los parametros ----------
        if not isinstance(tag, (str, int)):
            raise TypeError("Parameter 'tag' must be a string or an integer.")
        
        if not isinstance(pos, int) or pos < 1:
            raise ValueError("Parameter 'pos' must be an integer greater than or equal to 1.")

        #Toma los datos en un dict y los envia a "self._structure_layout["elements_layout"]"
        self._structure_layout["elements_layout"].append({  "tag": tag, 
                                                            "pos" : pos})
        
#+----------------------------------------------------------------+
#|                          THEME WIDGETS                         |
#+----------------------------------------------------------------+

#=========== ThemeWidgets ===========
"""
Da la orden del grid layout, maneja el orden de los widgets como si fueran una cuadricula
se puede modificar las cuadriculas indicando el tamaño de la fila o columna
los widgets adquieren el tamaño de las cuadriculas que fueron asignadas
"""
class ThemeWidgets():
    def __init__(self, widget_type: Literal["mvChildWindow", "mvImage", "mvImage", "mvImageButton"]):

        #---------- Validacion de datos de los parametros ----------
        if widget_type not in {"mvChildWindow", "mvImage", "mvImage", "mvImageButton"}:
            raise ValueError("Parameter 'widget_type' must be one of: 'mvChildWindow', 'mvImage', or 'mvImageButton'.")

        #Crea un tag unico
        self.tag = dpg.generate_uuid()

        #Crea la estructura de los temas widgets
        self.theme_w = {"widget_type": widget_type,
                        "id": self.tag,
                        "themes": []}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        else : 
            #Agrega los datos al submodulo theme
            theme.theme_data.add_theme(self.theme_w)

    def add_background_gradiant(self, *,
                                gradiant_type: Literal["Lineal_vertical", "Lineal_horizontal", "Radial"],
                                start_color: Tuple[int, int, int], 
                                end_color: Tuple[int, int, int],
                                rounded_corners: int = 0):

        #---------- Validacion de datos de los parametros ----------
        if gradiant_type not in {"Lineal_vertical", "Lineal_horizontal", "Radial"}:
            raise ValueError("Parameter 'gradiant_type' must be one of: 'Lineal_vertical', 'Lineal_horizontal', or 'Radial'.")

        if isinstance(start_color, tuple):
            if len(start_color) != 3 or not all(isinstance(val, int) for val in start_color):
                raise ValueError("Parameter 'color' must be a tuple of exactly 3 integers (R, G, B).")

        if isinstance(end_color, tuple):
            if len(end_color) != 3 or not all(isinstance(val, int) for val in end_color):
                raise ValueError("Parameter 'color' must be a tuple of exactly 3 integers (R, G, B).")
        
        if not isinstance(rounded_corners, int) or not 0 <= rounded_corners:
            raise ValueError("Parameter 'rounded_corners' must be a non-negative integer.")

        #Crea la estructura del tema
        structure = {   "theme": "background_gradiant",
                        "gradiant_type": gradiant_type,
                        "start_color": start_color,
                        "end_color": end_color,
                        "rounded_corners": rounded_corners}

        if self.theme_w["widget_type"] not in {"mvChildWindow"}:
            raise ValueError("ThemeWidgets.add_background_gradiant")

        #Es agregado a la lista de themes
        self.theme_w["themes"].append(structure)

    def add_background_color(self, *,
                            color: Tuple[int, int, int] = (0, 0, 0),
                            rounded_corners: int = 0):

        #---------- Validacion de datos de los parametros ----------
        if isinstance(color, tuple):
            if len(color) != 3 or not all(isinstance(val, int) for val in color):
                raise ValueError("Parameter 'color' must be a tuple of exactly 3 integers (R, G, B).")

        if not isinstance(rounded_corners, int) or not 0 <= rounded_corners:
            raise ValueError("Parameter 'rounded_corners' must be a non-negative integer.")

        #Crea la estructura del tema
        structure = {   "theme": "background_color",
                        "color": color,
                        "rounded_corners": rounded_corners}

        #Es agregado a la lista de themes
        self.theme_w["themes"].append(structure)

    def add_shadow( self, *, 
                    color: Tuple[int, int, int] = (0, 0, 0),
                    blur_level: int = 5,
                    opacity: float = 0.5,
                    thickness: int = 2,
                    pos: Tuple[int, int] = (0,4)):

        #---------- Validacion de datos de los parametros ----------
        if isinstance(color, tuple):
            if len(color) != 3 or not all(isinstance(val, int) for val in color):
                raise ValueError("Parameter 'color' must be a tuple of exactly 3 integers (R, G, B).")

        if not isinstance(blur_level, int) or not 1 <= blur_level:
            raise ValueError("Parameter 'blur_level' must be an integer greater than or equal to 1.")

        if not isinstance(opacity, float) or opacity < 0.0 or opacity > 1.0:
            raise ValueError("Parameter 'opacity' must be a float between 0.0 and 1.0 (inclusive).")

        if not isinstance(thickness, int) or not 0 <= thickness:
            raise ValueError("Parameter 'thickness' must be a non-negative integer.")
        
        if isinstance(pos, tuple):
            if len(pos) != 2 or not all(isinstance(val, int) for val in pos):
                raise ValueError("Parameter 'pos' must be a tuple of exactly 2 integers (x, y).")

        #Crea la estructura del tema
        structure = {   "theme": "shadow",
                        "color": color,
                        'blur_level': blur_level,
                        "opacity": opacity,
                        "thickness": thickness,
                        "pos": pos}

        #Es agregado a la lista de themes
        self.theme_w["themes"].append(structure)

#=========== bind_widget_theme ===========
"""
Esta funcion indica donde se deberia agregar los temas
IMPORTANTE:
Cada tema que creas contienen un tag propio
Se puede conseguir este tag al momento de asignarle un alias a un contexto de tema, Ejemplo:

with styleqs.ThemeWidgets(widget_type="mvChildWindow") as fondo:
    fondo.add_background_gradiant(  gradiant_type="Lineal_vertical", 
                                    start_color=(115, 250, 120), 
                                    end_color=(45, 75, 105))
tag_tema = fondo.tag    <------------------ Aqui se obtiene el tag

"""
def bind_widget_theme(tag: Union[str, int], tag_theme: int):

    #---------- Validacion de datos de los parametros ----------
    if not isinstance(tag, (str, int)):
        raise TypeError("Parameter 'tag' must be a string or an integer.")
    
    if not isinstance(tag_theme, int):
        raise TypeError("Parameter 'tag_theme' must be an integer that is created when creating a theme.")

    #Crea la estructura de los temas que seran asignados a los widgets
    structure = {"tag": tag, "tag_theme": tag_theme}

    #Es enviado a los datos del submodulo theme
    theme.theme_data.add_bind_theme(structure)

#+----------------------------------------------------------------+
#|                           ADD WIDGET                           |
#+----------------------------------------------------------------+
"""
Para poder agregarle temas a algunos widgets, se debe de tener unos parametros
pero dearpygui no proporciona estos parametros por lo cual se creo esta session
"""

#=========== add_image ===========
"""
Esta funcion agrega una imagen, se puede agregar de manera sencilla con el parametro route
el cual es la ruta de la imagen, adicional guarda en el user_data de dearpygui no de la funcion, 
datos necesarios para agregarle estilos al widget
"""
def add_image(*,route: str, 
                tag: Union[str, int] = 0, 
                width: int, 
                height: int, 
                user_data = None):

    #---------- Validacion de datos de los parametros ----------
    if not isinstance(route, str):
        raise TypeError("Parameter 'route' must be a string representing the image file path.")

    if not isinstance(tag, (str, int)):
        raise TypeError("Parameter 'tag' must be a string or an integer.")

    if not isinstance(width, int):
        raise TypeError("Parameter 'width' must be an integer.")
    if width <= 0:
        raise ValueError("Parameter 'width' must be a positive integer greater than zero.")

    if not isinstance(height, int):
        raise TypeError("Parameter 'height' must be an integer.")
    if height <= 0:
        raise ValueError("Parameter 'height' must be a positive integer greater than zero.")

    #Utiliza el submodulo add para registrar la imagen y obtener el tamaño de esta
    img, width_img, height_img = add.register_image(route)

    #Crea la imagen en dearpygui
    dpg.add_image(img, tag=tag, width=width, height=height, user_data={ "route": route,
                                                                        "size_img": (width_img, height_img),
                                                                        "user_data": user_data})

#=========== add_image_button ===========
#EN PROCESO...
"""
def add_image_button(*, route: str,
                        tag: str, 
                        width: int, 
                        height: int, 
                        user_data,
                        callback):
    pass
"""

#+----------------------------------------------------------------+
#|                               MORE                             |
#+----------------------------------------------------------------+
#Funciones importantes y necesarias para el uso de dearpygui y styleqs

#=========== clear_data ===========
#Funcion que elimina los datos(JSON - registro de imagenes y ZIP - imagenes guardadas)
def clear_data():
    #Variables que almacenan la ubicacion de los datos
    base_dir = Path(__file__).resolve().parent
    image_registration = base_dir / "data" / "image_registration.json"
    compressed_images = base_dir / "archives" / "compressed_images.zip"

    #Elimina el JSON
    if image_registration.exists():
        image_registration.unlink()

    #Elimina el ZIP
    if compressed_images.exists():
        compressed_images.unlink()

#=========== execute_after ===========
#Orden para ejecutar una funcion despues de que se ejecute todo el proceso de la funcion
def execute_after(callback):
    #Guarda la orden en los datos de more
    more.more_data.add_execute_after(callback)

#=========== copy_configuration ===========
#Funcion que copia el ancho y la posicion de un widget de otro widget
def copy_configuration(tag_widget, tag_widget_configuration):
    #Toma el tamaño y la posicion del widget a copiar
    width, height = dpg.get_item_rect_size(tag_widget_configuration)
    pos = dpg.get_item_pos(tag_widget_configuration)

    #Aplica el tamaño y la posicion
    dpg.configure_item(tag_widget, pos=pos, width=width, height=height)


import dearpygui.dearpygui as dpg
from typing import Literal, Union, List, Tuple

#==================== register_image ====================
#Funcion para registrar imagen con su ruta
def register_image(route):
    width, height, _, data = dpg.load_image(route)
    with dpg.texture_registry():
        img = dpg.add_static_texture(width, height, data)
    return img, width, height


from PIL import Image, ImageFilter, ImageDraw, ImageChops
from typing import Literal, Tuple
from pathlib import Path
from io import BytesIO
import zipfile
import shutil
import json

#---------- Variables con las ubicaciones de los zip y el JSON ----------
base_dir = Path(__file__).resolve().parent.parent
image_registration = base_dir / "data" / "image_registration.json"
compressed_images = base_dir / "archives" / "compressed_images.zip"
compressed_images_temp = base_dir / "archives" / "compressed_images_temp.zip"

#Si no existen crea las carpetas
image_registration.parent.mkdir(parents=True, exist_ok=True)
compressed_images.parent.mkdir(parents=True, exist_ok=True)
compressed_images_temp.parent.mkdir(parents=True, exist_ok=True)

#+----------------------------------------------------------------+
#|                     CLASE IMAGEVALIDATION                      |
#+----------------------------------------------------------------+

#=========== ImageValidation ===========
#Clase de validacion, valida los archivos zip con el JSON
class ImageValidation():
    def __init__(self):
        pass

    #==================== create_register ====================
    #Toma la estructura del tema y lo convierte en el formato que se utiliza en el JSON 
    def create_register(tag: str, theme: dict) -> dict:
        #Estructura del formato de JSON
        structure = {
            tag : {
                "core_effects" : [],
                "image_layer": [],
                "widget_type": theme["widget_type"]
            }
        }

        #Ubica los temas en su respectivo parametros
        for t in theme["themes"]:
            theme_type = t["theme"]
            match theme_type:
                case "background_gradiant":
                    structure[tag]["core_effects"].append(t)
                case "background_color":
                    structure[tag]["core_effects"].append(t)
                case "shadow":
                    structure[tag]["image_layer"].append(t)

        return structure

    #==================== create_register ====================
    #Toma la estructura del tema y lo convierte en el formato que se utiliza en el JSON 
    def remove_files_from_zip(files_to_remove):

        #---------- Abre el zip en modo lectura ----------
        with zipfile.ZipFile(compressed_images, 'r') as zip_read:
            #---------- Crea un zip temporal ----------
            with zipfile.ZipFile(compressed_images_temp, 'w') as zip_write:
                for item in zip_read.infolist():
                    # Si el archivo NO está en la lista a eliminar, lo copiamos
                    if item.filename not in files_to_remove:
                        zip_write.writestr(item, zip_read.read(item.filename))

        # Reemplazar el ZIP original con el nuevo
        shutil.move(compressed_images_temp, compressed_images)

    #==================== JSON_validation ====================
    """
    Sistema de validacion del JSON, elimina ordenes que ya no se llaman, 
    elimina imagenes que no se utilizaran y evita el rehacer imagenes que ya estan creadas
    """
    def JSON_validation(theme_widget, themes):
        content = {}
        #---------- Verifica si el archivo JSON existe ----------
        if image_registration.exists():
            #---------- Abre el zip en modo lectura ----------
            with open(image_registration, "r") as archive:
                try:
                    content = json.load(archive)
                except:
                    content = {}

                key_theme = list(theme_widget)[0]
                theme_widget = json.loads(json.dumps(theme_widget))

                #---------- Verifica si la orden ya esta en el Registro ----------
                if key_theme in content:
                    parameters_theme_widget = theme_widget[key_theme]
                    parameters_content = content[key_theme]

                    #---------- Verifica si no ha cambiado nada ----------
                    if parameters_theme_widget == parameters_content:
                        #ES IGUAL, NO HACER NADA
                        themes["themes"].clear()
                    #---------- Si cambio, verifica que cambio ----------
                    else:
                        #---------- Session donde verifica si cambio algo en el core_effects ----------
                        pt_tm_wg_core_effects = parameters_theme_widget["core_effects"]
                        pt_ct_core_effects = parameters_content["core_effects"]

                        if pt_tm_wg_core_effects != pt_ct_core_effects:
                            for element in pt_ct_core_effects:
                                if element in pt_tm_wg_core_effects:
                                    #ES IGUAL, REMUEVE LA ORDEN DE HACER LA IMAGEN
                                    themes["themes"].remove(element)
                                else:
                                    ImageValidation.remove_files_from_zip(f"{key_theme}/core.png")
                                    #NOTA: ELIMINAR IMAGEN RELACIONAD
                                    content[key_theme]["core_effects"].remove(element)

                            for element in pt_tm_wg_core_effects:
                                if element in pt_ct_core_effects:
                                    print(f"elemento existe en el JSON")
                                else:
                                    #NOTA: AGREGAR IMAGEN RELACIONADA
                                    content[key_theme]["core_effects"].append(element)

                        #---------- Session donde verifica si cambio algo en el image_layer ----------
                        pt_tm_wg_image_layer = parameters_theme_widget["image_layer"]
                        pt_ct_image_layer = parameters_content["image_layer"]

                        if pt_tm_wg_image_layer != pt_ct_image_layer:
                            for element in pt_ct_image_layer:
                                if element in pt_tm_wg_image_layer:
                                    #ES IGUAL, REMUEVE LA ORDEN DE HACER LA IMAGEN
                                    themes["themes"].remove(element)
                                else:
                                    ImageValidation.remove_files_from_zip(f"{key_theme}/image_layer/{element["theme"]}.png")
                                    #NOTA: ELIMINAR IMAGEN RELACIONADA
                                    content[key_theme]["image_layer"].remove(element)

                            for element in pt_tm_wg_image_layer:
                                if element in pt_ct_image_layer:
                                    print(f"elemento existe en el JSON")
                                else:
                                    #NOTA: AGREGAR IMAGEN RELACIONADA
                                    content[key_theme]["image_layer"].append(element)
                        with open(image_registration, "w") as archive:
                            json.dump(content, archive, indent=4)

                        return themes

                    with open(image_registration, "w") as archive:
                        json.dump(content, archive, indent=4)

                    return themes
                else:
                    content.update(theme_widget)

            with open(image_registration, "w") as archive:
                json.dump(content, archive, indent=4)

            return themes
        #---------- Si el archivo no existe lo crea ----------
        else:
            with open(image_registration, "w") as archive:
                content = theme_widget
                json.dump(content, archive, indent=4)
            return themes

    #==================== delete_keys ====================
    #Elimina del archivo JSON todas las claves que no estén en used_keys.
    def delete_keys(used_keys):
        #---------- Verifica si el archivo de registro de imágenes existe ----------
        if image_registration.exists():
            # Abre el archivo en modo lectura
            with open(image_registration, "r") as archivo:
                try:
                    content = json.load(archivo)
                except:
                    content = {}

        #Crea un nuevo diccionario solo con las claves que están en used_keys
        nuevas_keys = {k: v for k, v in content.items() if k in used_keys}

        #Escribe el nuevo JSON sin las keys no utilizadas
        with open(image_registration, "w") as archivo:
            json.dump(nuevas_keys, archivo, indent=4)

#+----------------------------------------------------------------+
#|                      CLASE IMAGE STYLEQS                       |
#+----------------------------------------------------------------+

#==================== ImageStyleQs ====================
#Clase que crea imgenes, agrega efectos y guarda en el archivo zip
class ImageStyleQs():
    def __init__(self):
        self._image = None

    #==================== load_image ====================
    #Carga imagen que esten el zip para asi aplicarle efectos
    def load_image(tag: str, img_type: Literal["core_effects", "image_layer"], img: str):
        #---------- ruta de la imagen ----------
        router = ""
        if img_type == "core_effects":
            router = f"{tag}/core.png"
        elif img_type == "image_layer":
            router = f"{tag}/{img_type}/{img}"
        
        #---------- Abre el zip en modo lectura ----------
        with zipfile.ZipFile(compressed_images, "r") as zipimg:
            #---------- Toma la imagen ----------
            with zipimg.open(router) as img_file:
                #carga la imagen de manera interna
                img_bytes = BytesIO(img_file.read())
                #Abre la imagen como pillow
                img = Image.open(img_bytes)

                return img

    #==================== CreateImage ====================
    #Crea las imagenes
    class CreateImage():
        def __init__(self):
            pass

        #==================== backgroud_color ====================
        #Hace la imagen de un color con el ancho y alto de los widgets
        def backgroud_color(width: int,
                            height: int,
                            color: Tuple[int, int, int, int]):

            size = (width, height)
            image = Image.new("RGBA", size, (0, 0, 0, 0))
            draw = ImageDraw.Draw(image)

            draw.rectangle((0,0, width, height), fill=color)
            return image

        #==================== background_gradiants ====================
        """
        Hace la imagen gradiante con el ancho y alto de los widgets, 
        tiene 3 tipos, lineal vertical, lineal horizontal y radial
        """
        def background_gradiants(gradiant_type: Literal["Lineal_vertical", "Lineal_horizontal", "Radial"],
                                width: int, height: int,
                                start_color: Tuple[int, int, int],
                                end_color: Tuple[int, int, int]):

            def lineal_vertical_gradiant():
                image = Image.new("RGB", (width, height))
                for y in range(height):
                    t = y / height

                    r = int(start_color[0] + (end_color[0] - start_color[0]) * t)
                    g = int(start_color[1] + (end_color[1] - start_color[1]) * t)
                    b = int(start_color[2] + (end_color[2] - start_color[2]) * t)

                    for x in range(width):
                        image.putpixel((x, y), (r, g, b))

                return image

            def lineal_horizontal_gradiant():
                image = Image.new("RGB", (width, height))
                for x in range(width):
                    t = x / height

                    r = int(start_color[0] + (end_color[0] - start_color[0]) * t)
                    g = int(start_color[1] + (end_color[1] - start_color[1]) * t)
                    b = int(start_color[2] + (end_color[2] - start_color[2]) * t)

                    for y in range(height):
                        image.putpixel((x, y), (r, g, b))

                return image

            def radial_gradient():
                from math import hypot
                image = Image.new("RGB", (width, height))
                center_x = width // 2
                center_y = height // 2
                max_dist = hypot(center_x, center_y)

                for y in range(height):
                    for x in range(width):

                        dist = hypot(x - center_x, y - center_y)
                        t = dist / max_dist
                        t = min(max(t, 0), 1)

                        r = int(start_color[0] + (end_color[0] - start_color[0]) * t)
                        g = int(start_color[1] + (end_color[1] - start_color[1]) * t)
                        b = int(start_color[2] + (end_color[2] - start_color[2]) * t)

                        image.putpixel((x, y), (r, g, b))

                return image

            match gradiant_type:
                case "Lineal_vertical":
                    return lineal_vertical_gradiant()
                case "Lineal_horizontal":
                    return lineal_horizontal_gradiant()
                case "Radial":
                    return radial_gradient()

    #==================== EffectsImage ====================
    #Agrega efectos a las imagenes
    class EffectsImage():
        def __init__(self, image):
            self.image = image

        #==================== silhouette_fill ====================
        #Recrea la silueta de una imagen png
        def silhouette_fill(self,color: Tuple[int, int, int]):
            self.image = self.image.convert("RGBA")
            pixels = self.image.load()

            for y in range(self.image.height):
                for x in range(self.image.width):
                    r, b, g, a = pixels[x, y]
                    if a > 0:
                        pixels[x, y] = (*color, a)
            
            return self

        #==================== silhouette_fill ====================
        #Recrea contorno
        def outline(self,color: Tuple[int, int, int], thickness: int = 2):
            margin = thickness
            width, height = self.image.size
            new_size = (width + margin * 2, height + margin * 2)

            expanded_image = Image.new("RGBA", new_size, (0, 0, 0, 0))
            expanded_image.paste(self.image, (margin, margin))
            self.image = expanded_image

            alpha = self.image.split()[-1]
            mask = alpha.point(lambda p: 255 if p > 0 else 0).convert("L")
            expanded = mask.filter(ImageFilter.MaxFilter(thickness * 2 + 1))
            border_mask = ImageChops.subtract(expanded, mask)

            outline_image = Image.new("RGBA", self.image.size, (0, 0, 0, 0))
            outline_pixels = outline_image.load()
            border_pixels = border_mask.load()

            for y in range(self.image.height):
                for x in range(self.image.width):
                    if border_pixels[x, y] > 0:
                        outline_pixels[x, y] = (*color, 255)

            self.image = Image.alpha_composite(outline_image, self.image)
            return self

        #==================== blur ====================
        #Agrega efecto de desenfoque
        def blur(self,radius: int):
            margin = radius * radius
            width, height = self.image.size
            new_size = (width + margin * 2, height + margin * 2)

            expanded_image = Image.new("RGBA", new_size, (0, 0, 0, 0))
            expanded_image.paste(self.image, (margin, margin))
            self.image = expanded_image

            self.image = self.image.filter(ImageFilter.GaussianBlur(radius=radius))
            return self

        #==================== rounded_corners ====================
        #Redondea las esquinas de las images
        def rounded_corners(self, radius: int):
            width, height = self.image.size

            mask = Image.new("L", (width, height), 0)
            draw = ImageDraw.Draw(mask)
            draw.rounded_rectangle([(0, 0), (width, height)], radius=radius, fill=255)

            self.image.putalpha(mask)
            return self

        #==================== opacity ====================
        #Baja la opacidad de las imagenes
        def opacity(self,opacity: float):
            self.image = self.image.convert("RGBA")
            r, g, b, a = self.image.split()
            a = a.point(lambda p: int(p * opacity))
            self.image = Image.merge("RGBA", (r, b, g, a))
            return self

        #==================== image_return ====================
        #Retorna la imagen de pillow para que despues pueda ser guardada
        def image_return(self):
            return self.image

    #==================== show ====================
    #Muestra la imagen (util para pruebas)
    def show(image):
        image.show()

    #==================== save ====================
    #Guarda las imagenes en el zip
    def save(image, tag: str, site: Literal["core_effects", "image_layer"], img_type: str = ""):
        #---------- Comprueba si el zip existe ----------
        if compressed_images.exists():
            #---------- Abre el zip en modo append(agregar o añadir) ----------
            with zipfile.ZipFile(compressed_images, mode="a") as zipimg:
                #Utilizado para guarda la imagen en memoria
                buffer = BytesIO()
                #Guarda la imagen
                image.save(buffer, format="PNG")
                buffer.seek(0)

                #Ruta para guardar la imagen en el zip
                archive_site = ""
                if site == "core_effects":
                    archive_site = f"{tag}/core.png"
                elif site == "image_layer":
                    archive_site = f"{tag}/image_layer/{img_type}.png"

                #Guarda la imagen en el zip
                zipimg.writestr(archive_site , buffer.read())

        #---------- Si no existe la crea ----------
        else:
            #---------- Abre el zip en modo write(escribir o crear zip) ----------
            with zipfile.ZipFile(compressed_images, "w", zipfile.ZIP_DEFLATED) as zipimg:
                #Utilizado para guarda la imagen en memoria
                buffer = BytesIO()
                #Guarda la imagen
                image.save(buffer, format="PNG")
                buffer.seek(0)

                #Ruta para guardar la imagen en el zip
                archive_site = ""
                if site == "core_effects":
                    archive_site = f"{tag}/core.png"
                elif site == "image_layer":
                    archive_site = f"{tag}/image_layer/{img_type}.png"

                #Guarda la imagen en el zip
                zipimg.writestr(archive_site , buffer.read())

"""
prueba = ImageStyleQs.CreateImage.frame_backgroud_color(width=200, height=200, color=(255,0,0,255))
prueba = ImageStyleQs.EffectsImage(prueba)\
    .silhouette_fill(color=(0,0,0))\
    .rounded_corners(100)\
    .outline(color=(255,255,255), thickness=2)\
    .blur(10)\
    .image_return()
ImageStyleQs.save(prueba, "Nose", "image_layer", "shadow")
"""